"""Interactions with GitHub."""

from __future__ import annotations

from collections.abc import Iterable
from functools import partial
from typing import Any, Literal, cast

import requests
from ratelimit import rate_limited
from tqdm import tqdm

from .config_api import get_config
from .constants import TimeoutHint

__all__ = [
    "MAXIMUM_SEARCH_PAGE_SIZE",
    "delete_branch",
    "get_contributions",
    "get_default_branch",
    "get_issue",
    "get_issues",
    "get_pull_requests",
    "get_repository",
    "get_repository_commit_activity",
    "get_token",
    "get_topics",
    "get_user_events",
    "post_pull",
    "requests_get_github",
    "requests_post_github",
    "search_code",
]


def get_token(*, passthrough: str | None = None, raise_on_missing: bool = False) -> str | None:
    """Get a GitHub token."""
    rv = get_config("github", "token", passthrough=passthrough, raise_on_missing=raise_on_missing)
    return rv  # type:ignore


def get_headers(
    token: str | None = None, *, raise_on_missing: bool = False, preview: bool = False
) -> dict[str, str]:
    """Get GitHub headers."""
    headers = {
        "Accept": "application/vnd.github.mercy-preview+json"
        if preview
        else "application/vnd.github.v3+json",
    }
    token = get_token(passthrough=token, raise_on_missing=raise_on_missing)
    if token:
        headers["Authorization"] = f"token {token}"
    return headers


def requests_get_github(path: str, **kwargs: Any) -> requests.Response:
    """Make a GET request to the GitHub API."""
    return _request_github("GET", path, **kwargs)


def requests_post_github(path: str, **kwargs: Any) -> requests.Response:
    """Make a POST request to the GitHub API."""
    return _request_github("POST", path, **kwargs)


def requests_delete_github(path: str, **kwargs: Any) -> requests.Response:
    """Make a DELETE request to the GitHub API."""
    return _request_github("DELETE", path, **kwargs)


@rate_limited(calls=5_000, period=60 * 60)
def _request_github(
    method: Literal["GET", "POST", "DELETE"],
    path: str,
    params: dict[str, Any] | None = None,
    token: str | None = None,
    timeout: TimeoutHint = None,
    require_token: bool = False,
    preview: bool = False,
    **kwargs: Any,
) -> requests.Response:
    """Make a POST request to the GitHub API."""
    path = path.lstrip("/")
    url = f"https://api.github.com/{path}"
    headers = get_headers(token=token, raise_on_missing=require_token, preview=preview)
    return requests.request(method, url, headers=headers, params=params, timeout=timeout, **kwargs)


def post_pull(
    owner: str,
    repo: str,
    *,
    title: str | None = None,
    head: str,
    head_repo: str | None = None,
    base: str,
    body: str | None = None,
    maintainer_can_modify: bool = True,
    draft: bool = False,
    issue: int | None = None,
    **kwargs: Any,
) -> requests.Response:
    """Post a pull request."""
    # https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#create-a-pull-request
    if title is None and issue is None:
        raise ValueError("title and issue can't both be none")
    if "params" in kwargs:
        raise ValueError("should pass arguments directly")
    return requests_post_github(
        f"repos/{owner}/{repo}/pulls",
        params={
            "title": title,
            "head": head,
            "head_repo": head_repo,
            "base": base,
            "body": body,
            "maintainer_can_modify": maintainer_can_modify,
            "draft": draft,
            "issue": issue,
        },
        **kwargs,
    )


def delete_branch(owner: str, repo: str, branch_name: str, **kwargs: Any) -> requests.Response:
    """Delete a branch.

    :param owner: The repository owner
    :param repo: The repository name
    :param branch_name: The branch name or head ref
    :returns: A response from the GitHub API after running the (secret) delete endpoint
    """
    # see https://github.com/orgs/community/discussions/24603
    return requests_delete_github(f"repos/{owner}/{repo}/git/refs/heads/{branch_name}", **kwargs)


def get_repository(owner: str, repo: str, **kwargs: Any) -> requests.Response:
    """Get information about a repository."""
    return requests_get_github(f"repos/{owner}/{repo}", **kwargs)


def get_default_branch(owner: str, repo: str, **kwargs: Any) -> str:
    """Get the default branch for the repository."""
    res = get_repository(owner, repo, **kwargs).json()
    return cast(str, res["default_branch"])


def get_issues(owner: str, repo: str, **kwargs: Any) -> requests.Response:
    """Get issues from a repository."""
    return requests_get_github(f"repos/{owner}/{repo}/issues", **kwargs)


def get_issue(owner: str, repo: str, issue: int, **kwargs: Any) -> requests.Response:
    """Get an issue from a repository."""
    return requests_get_github(f"repos/{owner}/{repo}/issues/{issue}", **kwargs)


def get_pull_requests(owner: str, repo: str, **kwargs: Any) -> requests.Response:
    """Get pull requests from a repository."""
    return requests_get_github(f"repos/{owner}/{repo}/pulls", **kwargs)


def get_user_events(user: str) -> requests.Response:
    """Get events for a user."""
    return requests_get_github(f"users/{user}/events")


def get_contributions(owner: str, repo: str, **kwargs: Any) -> requests.Response:
    """Get contributors to a repository."""
    return requests_get_github(f"repos/{owner}/{repo}/stats/contributors", **kwargs)


def get_repository_commit_activity(owner: str, repo: str, **kwargs: Any) -> requests.Response:
    """Get commit activity to a repository."""
    return requests_get_github(f"repos/{owner}/{repo}/stats/commit_activity", **kwargs)


def get_topics(owner: str, repo: str, *, preview: bool = True, **kwargs: Any) -> requests.Response:
    """Get topics from the repository."""
    return requests_get_github(f"repos/{owner}/{repo}/topics", preview=preview, **kwargs)


#: Maximum number of records per page in code search
MAXIMUM_SEARCH_PAGE_SIZE = 40


def search_code(
    query: str,
    *,
    page_size: int | None = None,
    progress: bool = True,
    inner_progress: bool = True,
) -> Iterable[dict[str, Any]]:
    """Search GitHub code."""
    page = 1
    if page_size is None:
        page_size = MAXIMUM_SEARCH_PAGE_SIZE
    if page_size > MAXIMUM_SEARCH_PAGE_SIZE:
        page_size = MAXIMUM_SEARCH_PAGE_SIZE

    inner_tqdm = partial(tqdm, disable=not inner_progress, unit="record", leave=False)

    initial_res = _search_code_helper(page_size=page_size, page=page, query=query)
    initial_res.raise_for_status()
    initial_res_json = initial_res.json()
    total = initial_res_json["total_count"]
    total_pages = 1 + (total // page_size)
    yield from inner_tqdm(initial_res_json["items"], desc="Page 1")

    with tqdm(
        total=total_pages, unit="page", disable=not progress, desc="Paginating code search results"
    ) as tbar:
        tbar.update(1)

        while page_size * page < total:
            page += 1
            tbar.update(1)
            res = _search_code_helper(page=page, page_size=page_size, query=query)
            res.raise_for_status()
            res_json = res.json()
            yield from inner_tqdm(res_json["items"], desc=f"Page {page}")


def _search_code_helper(page_size: int, page: int, query: str) -> requests.Response:
    return requests_get_github(
        "search/code",
        params={
            "per_page": page_size,
            "page": page,
            "sort": "indexed",
            "q": query,
        },
    )



"""Utilities."""

from __future__ import annotations

import contextlib
import csv
import gzip
import hashlib
import io
import logging
import lzma
import os
import pickle
import shutil
import tarfile
import tempfile
import typing
import urllib.error
import zipfile
from collections.abc import Collection, Generator, Iterable, Iterator, Mapping
from functools import partial
from io import BytesIO, StringIO
from pathlib import Path, PurePosixPath
from subprocess import check_output
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, TextIO, TypeAlias, cast
from urllib.parse import urlparse
from urllib.request import urlretrieve
from uuid import uuid4

import requests
from tqdm.auto import tqdm

from .constants import (
    PYSTOW_HOME_ENVVAR,
    PYSTOW_NAME_DEFAULT,
    PYSTOW_NAME_ENVVAR,
    PYSTOW_USE_APPDIRS,
    README_TEXT,
    TimeoutHint,
)

if TYPE_CHECKING:
    import _csv

    import botocore.client
    import bs4
    import lxml.etree
    import numpy.typing
    import pandas
    import rdflib

__all__ = [
    "MODE_MAP",
    "OPERATION_VALUES",
    "REPRESENTATION_VALUES",
    "REVERSE_MODE_MAP",
    "DownloadBackend",
    "DownloadError",
    "Hash",
    "HexDigestError",
    "HexDigestMismatch",
    "InvalidOperationError",
    "InvalidRepresentationError",
    "Operation",
    "Representation",
    "UnexpectedDirectory",
    "UnexpectedDirectoryError",
    "download",
    "download_from_google",
    "download_from_s3",
    "get_base",
    "get_commit",
    "get_df_io",
    "get_hash_hexdigest",
    "get_hashes",
    "get_hexdigests_remote",
    "get_home",
    "get_mode_pair",
    "get_name",
    "get_np_io",
    "get_offending_hexdigests",
    "get_soup",
    "getenv_path",
    "gunzip",
    "mkdir",
    "mock_envvar",
    "mock_home",
    "n",
    "name_from_s3_key",
    "name_from_url",
    "open_tarfile",
    "open_zip_reader",
    "open_zip_writer",
    "open_zipfile",
    "path_to_sqlite",
    "raise_on_digest_mismatch",
    "read_rdf",
    "read_tarfile_csv",
    "read_tarfile_xml",
    "read_zip_np",
    "read_zipfile_csv",
    "read_zipfile_rdf",
    "read_zipfile_xml",
    "safe_open",
    "safe_open_dict_reader",
    "safe_open_dict_writer",
    "safe_open_reader",
    "safe_open_writer",
    "write_lzma_csv",
    "write_pickle_gz",
    "write_tarfile_csv",
    "write_tarfile_xml",
    "write_zipfile_csv",
    "write_zipfile_np",
    "write_zipfile_rdf",
    "write_zipfile_xml",
]

logger = logging.getLogger(__name__)

#: Represents an available backend for downloading
DownloadBackend: TypeAlias = Literal["urllib", "requests"]

#: This type alias uses a stub-only constructor, meaning that
#: hashlib._Hash isn't actually part of the code, but MyPy injects it
#: so we can do type checking
Hash: TypeAlias = "hashlib._Hash"

Reader: TypeAlias = "_csv._reader"
Writer: TypeAlias = "_csv._writer"

#: A human-readable flag for how to open a file.
Operation: TypeAlias = Literal["read", "write"]
OPERATION_VALUES: set[str] = set(typing.get_args(Operation))

#: A human-readable flag for how to open a file.
Representation: TypeAlias = Literal["text", "binary"]
REPRESENTATION_VALUES: set[str] = set(typing.get_args(Representation))

#: Characters for "unqualified" modes, which might be interpreted
#: differently by different functions
UnqualifiedMode: TypeAlias = Literal["r", "w"]

#: Characters for "qualified" modes, which are absolute (as opposed to
#: :data:`UnqualifiedMode`, which is context-dependent)
QualifiedMode: TypeAlias = Literal["rt", "wt", "rb", "wb"]

ModePair: TypeAlias = tuple[Operation, Representation]

#: A mapping between operation/representation pairs and qualified modes
MODE_MAP: dict[ModePair, QualifiedMode] = {
    ("read", "text"): "rt",
    ("read", "binary"): "rb",
    ("write", "text"): "wt",
    ("write", "binary"): "wb",
}

#: A mapping between qualified modes and operation/representation pairs
REVERSE_MODE_MAP: dict[QualifiedMode, ModePair] = {
    "rt": ("read", "text"),
    "rb": ("read", "binary"),
    "wt": ("write", "text"),
    "wb": ("write", "binary"),
}

UNQUALIFIED_TEXT_MAP: dict[UnqualifiedMode, ModePair] = {
    "r": ("read", "text"),
    "w": ("write", "text"),
}
UNQUALIFIED_BINARY_MAP: dict[UnqualifiedMode, ModePair] = {
    "r": ("read", "binary"),
    "w": ("write", "binary"),
}


def get_mode_pair(
    mode: UnqualifiedMode | QualifiedMode, interpretation: Representation
) -> ModePair:
    """Get the mode pair."""
    match mode:
        case "rt" | "wt" | "rb" | "wb":
            return REVERSE_MODE_MAP[mode]
        case "r" | "w" if interpretation == "text":
            return UNQUALIFIED_TEXT_MAP[mode]
        case "r" | "w" if interpretation == "binary":
            return UNQUALIFIED_BINARY_MAP[mode]
        case _:
            raise ValueError(f"invalid mode: {mode}")


class HexDigestMismatch(NamedTuple):
    """Contains information about a hexdigest mismatch."""

    #: the name of the algorithm
    name: str
    #: the observed/actual hexdigest, encoded as a string
    actual: str
    #: the expected hexdigest, encoded as a string
    expected: str


class HexDigestError(ValueError):
    """Thrown if the hashsums do not match expected hashsums."""

    def __init__(self, offending_hexdigests: Collection[HexDigestMismatch]):
        """Instantiate the exception.

        :param offending_hexdigests: The result from :func:`get_offending_hexdigests`
        """
        self.offending_hexdigests = offending_hexdigests

    def __str__(self) -> str:
        return "\n".join(
            (
                "Hexdigest of downloaded file does not match the expected ones!",
                *(
                    f"\t{name} actual: {actual} vs. expected: {expected}"
                    for name, actual, expected in self.offending_hexdigests
                ),
            )
        )


class UnexpectedDirectoryError(FileExistsError):
    """Thrown if a directory path is given where file path should have been."""

    def __init__(self, path: Path):
        """Instantiate the exception.

        :param path: The path to a directory that should have been a file.
        """
        self.path = path

    def __str__(self) -> str:
        return f"got directory instead of file: {self.path}"


#: Backwards compatible name
UnexpectedDirectory = UnexpectedDirectoryError


def get_hexdigests_remote(
    hexdigests_remote: Mapping[str, str] | None, hexdigests_strict: bool = False
) -> Mapping[str, str]:
    """Process hexdigests via URLs.

    :param hexdigests_remote: The expected hexdigests as (algorithm_name, url to file
        with expected hex digest) pairs.
    :param hexdigests_strict: Set this to `False` to stop automatically checking for the
        `algorithm(filename)=hash` format

    :returns: A mapping of algorithms to hexdigests
    """
    rv = {}
    for key, url in (hexdigests_remote or {}).items():
        text = requests.get(url, timeout=15).text
        if not hexdigests_strict and "=" in text:
            text = text.rsplit("=", 1)[-1].strip()
        rv[key] = text
    return rv


def get_offending_hexdigests(
    path: str | Path,
    chunk_size: int | None = None,
    hexdigests: Mapping[str, str] | None = None,
    hexdigests_remote: Mapping[str, str] | None = None,
    hexdigests_strict: bool = False,
) -> Collection[HexDigestMismatch]:
    """Check a file for hash sums.

    :param path: The file path.
    :param chunk_size: The chunk size for reading the file.
    :param hexdigests: The expected hexdigests as (algorithm_name, expected_hex_digest)
        pairs.
    :param hexdigests_remote: The expected hexdigests as (algorithm_name, url to file
        with expected hexdigest) pairs.
    :param hexdigests_strict: Set this to false to stop automatically checking for the
        `algorithm(filename)=hash` format

    :returns: A collection of observed / expected hexdigests where the digests do not
        match.
    """
    hexdigests = dict(
        **(hexdigests or {}),
        **get_hexdigests_remote(hexdigests_remote, hexdigests_strict=hexdigests_strict),
    )

    # If there aren't any keys in the combine dictionaries,
    # then there won't be any mismatches
    if not hexdigests:
        return []

    logger.info(f"Checking hash sums for file: {path}")

    # instantiate algorithms
    algorithms = get_hashes(path=path, names=set(hexdigests), chunk_size=chunk_size)

    # Compare digests
    mismatches = []
    for alg, expected_digest in hexdigests.items():
        observed_digest = algorithms[alg].hexdigest()
        if observed_digest != expected_digest:
            logger.error(f"{alg} expected {expected_digest} but got {observed_digest}.")
            mismatches.append(HexDigestMismatch(alg, observed_digest, expected_digest))
        else:
            logger.debug(f"Successfully checked with {alg}.")

    return mismatches


def get_hashes(
    path: str | Path,
    names: Iterable[str],
    *,
    chunk_size: int | None = None,
) -> Mapping[str, Hash]:
    """Calculate several hexdigests of hash algorithms for a file concurrently.

    :param path: The file path.
    :param names: Names of the hash algorithms in :mod:`hashlib`
    :param chunk_size: The chunk size for reading the file.

    :returns: A collection of observed hexdigests
    """
    path = Path(path).resolve()
    if chunk_size is None:
        chunk_size = 64 * 2**10

    # instantiate hash algorithms
    algorithms: Mapping[str, Hash] = {name: hashlib.new(name) for name in names}

    # calculate hash sums of file incrementally
    buffer = memoryview(bytearray(chunk_size))
    with path.open("rb", buffering=0) as file:
        for this_chunk_size in iter(lambda: file.readinto(buffer), 0):
            for alg in algorithms.values():
                alg.update(buffer[:this_chunk_size])

    return algorithms


def get_hash_hexdigest(
    path: str | Path,
    name: str,
    *,
    chunk_size: int | None = None,
) -> str:
    """Get a hash digest for a single hash."""
    r = get_hashes(path, [name], chunk_size=chunk_size)
    return r[name].hexdigest()


def raise_on_digest_mismatch(
    *,
    path: Path,
    hexdigests: Mapping[str, str] | None = None,
    hexdigests_remote: Mapping[str, str] | None = None,
    hexdigests_strict: bool = False,
) -> None:
    """Raise a HexDigestError if the digests do not match.

    :param path: The file path.
    :param hexdigests: The expected hexdigests as (algorithm_name, expected_hex_digest)
        pairs.
    :param hexdigests_remote: The expected hexdigests as (algorithm_name, url to file
        with expected hexdigest) pairs.
    :param hexdigests_strict: Set this to false to stop automatically checking for the
        `algorithm(filename)=hash` format

    :raises HexDigestError: if there are any offending hex digests The expected
        hexdigests as (algorithm_name, url to file with expected hexdigest) pairs.
    """
    offending_hexdigests = get_offending_hexdigests(
        path=path,
        hexdigests=hexdigests,
        hexdigests_remote=hexdigests_remote,
        hexdigests_strict=hexdigests_strict,
    )
    if offending_hexdigests:
        raise HexDigestError(offending_hexdigests)


class TqdmReportHook(tqdm):  # type:ignore
    """A custom progress bar that can be used with urllib.

    Based on https://gist.github.com/leimao/37ff6e990b3226c2c9670a2cd1e4a6f5
    """

    def update_to(
        self,
        blocks: int = 1,
        block_size: int = 1,
        total_size: int | None = None,
    ) -> None:
        """Update the internal state based on a urllib report hook.

        :param blocks: Number of blocks transferred so far
        :param block_size: Size of each block (in tqdm units)
        :param total_size: Total size (in tqdm units). If [default: None] remains
            unchanged.
        """
        if total_size is not None:
            self.total = total_size
        self.update(blocks * block_size - self.n)  # will also set self.n = b * bsize


def download(
    url: str,
    path: str | Path,
    force: bool = True,
    clean_on_failure: bool = True,
    backend: DownloadBackend = "urllib",
    hexdigests: Mapping[str, str] | None = None,
    hexdigests_remote: Mapping[str, str] | None = None,
    hexdigests_strict: bool = False,
    progress_bar: bool = True,
    tqdm_kwargs: Mapping[str, Any] | None = None,
    **kwargs: Any,
) -> None:
    """Download a file from a given URL.

    :param url: URL to download
    :param path: Path to download the file to
    :param force: If false and the file already exists, will not re-download.
    :param clean_on_failure: If true, will delete the file on any exception raised
        during download
    :param backend: The downloader to use. Choose 'urllib' or 'requests'
    :param hexdigests: The expected hexdigests as (algorithm_name, expected_hex_digest)
        pairs.
    :param hexdigests_remote: The expected hexdigests as (algorithm_name, url to file
        with expected hexdigest) pairs.
    :param hexdigests_strict: Set this to false to stop automatically checking for the
        `algorithm(filename)=hash` format
    :param progress_bar: Set to true to show a progress bar while downloading
    :param tqdm_kwargs: Override the default arguments passed to :class:`tadm.tqdm` when
        progress_bar is True.
    :param kwargs: The keyword arguments to pass to :func:`urllib.request.urlretrieve`
        or to `requests.get` depending on the backend chosen. If using 'requests'
        backend, `stream` is set to True by default.

    :raises Exception: Thrown if an error besides a keyboard interrupt is thrown during
        download
    :raises KeyboardInterrupt: If a keyboard interrupt is thrown during download
    :raises UnexpectedDirectory: If a directory is given for the ``path`` argument
    :raises ValueError: If an invalid backend is chosen
    :raises DownloadError: If an error occurs during download
    """
    path = Path(path).resolve()

    if path.is_dir():
        raise UnexpectedDirectoryError(path)
    if path.is_file() and not force:
        raise_on_digest_mismatch(
            path=path,
            hexdigests=hexdigests,
            hexdigests_remote=hexdigests_remote,
            hexdigests_strict=hexdigests_strict,
        )
        logger.debug("did not re-download %s from %s", path, url)
        return

    _tqdm_kwargs = {
        "unit": "B",
        "unit_scale": True,
        "unit_divisor": 1024,
        "miniters": 1,
        "disable": not progress_bar,
        "desc": f"Downloading {path.name}",
        "leave": False,
    }
    if tqdm_kwargs:
        _tqdm_kwargs.update(tqdm_kwargs)

    try:
        if backend == "urllib":
            logger.info("downloading with urllib from %s to %s", url, path)
            with TqdmReportHook(**_tqdm_kwargs) as t:
                try:
                    urlretrieve(url, path, reporthook=t.update_to, **kwargs)  # noqa:S310
                except urllib.error.URLError as e:
                    raise DownloadError(backend, url, path, e) from e
        elif backend == "requests":
            kwargs.setdefault("stream", True)
            try:
                # see https://requests.readthedocs.io/en/master/user/quickstart/#raw-response-content
                # pattern from https://stackoverflow.com/a/39217788/5775947
                with requests.get(url, **kwargs) as response, path.open("wb") as file:  # noqa:S113
                    logger.info(
                        "downloading (stream=%s) with requests from %s to %s",
                        kwargs["stream"],
                        url,
                        path,
                    )
                    # Solution for progress bar from https://stackoverflow.com/a/63831344/5775947
                    total_size = int(response.headers.get("Content-Length", 0))
                    # Decompress if needed
                    response.raw.read = partial(  # type:ignore[method-assign]
                        response.raw.read, decode_content=True
                    )
                    with tqdm.wrapattr(
                        response.raw, "read", total=total_size, **_tqdm_kwargs
                    ) as fsrc:
                        shutil.copyfileobj(fsrc, file)
            except requests.exceptions.ConnectionError as e:
                raise DownloadError(backend, url, path, e) from e
        else:
            raise ValueError(f'Invalid backend: {backend}. Use "requests" or "urllib".')
    except (Exception, KeyboardInterrupt):
        if clean_on_failure:
            _unlink(path)
        raise

    raise_on_digest_mismatch(
        path=path,
        hexdigests=hexdigests,
        hexdigests_remote=hexdigests_remote,
        hexdigests_strict=hexdigests_strict,
    )


class DownloadError(OSError):
    """An error that wraps information from a requests or urllib download failure."""

    def __init__(
        self,
        backend: DownloadBackend,
        url: str,
        path: Path,
        exc: urllib.error.URLError | requests.exceptions.ConnectionError,
    ) -> None:
        """Initialize the error.

        :param backend: The backend used
        :param url: The url that failed to download
        :param path: The path that was supposed to be downloaded to
        :param exc: The exception raised
        """
        self.backend = backend
        self.url = url
        self.path = path
        self.exc = exc
        # TODO parse out HTTP error code, if possible

    def __str__(self) -> str:
        return f"Failed with {self.backend} to download {self.url} to {self.path}"


def name_from_url(url: str) -> str:
    """Get the filename from the end of the URL.

    :param url: A URL

    :returns: The name of the file at the end of the URL
    """
    parse_result = urlparse(url)
    path = PurePosixPath(parse_result.path)
    name = path.name
    return name


def base_from_gzip_name(name: str) -> str:
    """Get the base name for a file after stripping the gz ending.

    :param name: The name of the gz file

    :returns: The cleaned name of the file, with no gz ending

    :raises ValueError: if the file does not end with ".gz"
    """
    if not name.endswith(".gz"):
        raise ValueError(f"Name does not end with .gz: {name}")
    return name[: -len(".gz")]


def name_from_s3_key(key: str) -> str:
    """Get the filename from the S3 key.

    :param key: A S3 path

    :returns: The name of the file
    """
    return key.split("/")[-1]


def mkdir(path: Path, ensure_exists: bool = True) -> None:
    """Make a directory (or parent directory if a file is given) if flagged with ``ensure_exists``.

    :param path: The path to a directory
    :param ensure_exists: Should the directories leading to the path be created if they
        don't already exist?
    """
    if ensure_exists:
        path.mkdir(exist_ok=True, parents=True)


@contextlib.contextmanager
def mock_envvar(envvar: str, value: str) -> Iterator[None]:
    """Mock the environment variable then delete it after the test is over.

    :param envvar: The environment variable to mock
    :param value: The value to temporarily put in the environment variable during this
        mock.

    :yield: None, since this just mocks the environment variable for the time being.
    """
    original_value = os.environ.get(envvar)
    os.environ[envvar] = value
    yield
    if original_value is None:
        del os.environ[envvar]
    else:
        os.environ[envvar] = original_value


@contextlib.contextmanager
def mock_home() -> Iterator[Path]:
    """Mock the PyStow home environment variable, yields the directory name.

    :yield: The path to the temporary directory.
    """
    with tempfile.TemporaryDirectory() as directory:
        with mock_envvar(PYSTOW_HOME_ENVVAR, directory):
            yield Path(directory)


def getenv_path(envvar: str, default: Path, ensure_exists: bool = True) -> Path:
    """Get an environment variable representing a path, or use the default.

    :param envvar: The environmental variable name to check
    :param default: The default path to return if the environmental variable is not set
    :param ensure_exists: Should the directories leading to the path be created if they
        don't already exist?

    :returns: A path either specified by the environmental variable or by the default.
    """
    rv = Path(os.getenv(envvar, default=default)).expanduser()
    mkdir(rv, ensure_exists=ensure_exists)
    return rv


def n() -> str:
    """Get a random string for testing.

    :returns: A random string for testing purposes.
    """
    return str(uuid4())


def get_df_io(df: pandas.DataFrame, sep: str = "\t", index: bool = False, **kwargs: Any) -> BytesIO:
    """Get the dataframe as bytes.

    :param df: A dataframe
    :param sep: The separator in the dataframe. Overrides Pandas default to use a tab.
    :param index: Should the index be output? Overrides the Pandas default to be false.
    :param kwargs: Additional kwargs to pass to :func:`pandas.DataFrame.to_csv`.

    :returns: A bytes object that can be used as a file.
    """
    sio = StringIO()
    df.to_csv(sio, sep=sep, index=index, **kwargs)
    sio.seek(0)
    bio = BytesIO(sio.read().encode("utf-8"))
    return bio


def get_np_io(arr: numpy.typing.ArrayLike, **kwargs: Any) -> BytesIO:
    """Get the numpy object as bytes.

    :param arr: Array-like
    :param kwargs: Additional kwargs to pass to :func:`numpy.save`.

    :returns: A bytes object that can be used as a file.
    """
    import numpy as np

    bio = BytesIO()
    np.save(bio, arr, **kwargs)
    bio.seek(0)
    return bio


def write_pickle_gz(
    obj: Any,
    path: str | Path,
    **kwargs: Any,
) -> None:
    """Write an object to a gzipped pickle.

    :param obj: The object to write
    :param path: The path of the file to write to
    :param kwargs: Additional kwargs to pass to :func:`pickle.dump`
    """
    with gzip.open(path, mode="wb") as file:
        pickle.dump(obj, file, **kwargs)


def write_lzma_csv(
    df: pandas.DataFrame,
    path: str | Path,
    sep: str = "\t",
    index: bool = False,
    **kwargs: Any,
) -> None:
    """Write a dataframe as an lzma-compressed file.

    :param df: A dataframe
    :param path: The path to the resulting LZMA compressed dataframe file
    :param sep: The separator in the dataframe. Overrides Pandas default to use a tab.
    :param index: Should the index be output? Overrides the Pandas default to be false.
    :param kwargs: Additional kwargs to pass to :func:`get_df_io` and transitively to
        :func:`pandas.DataFrame.to_csv`.
    """
    bytes_io = get_df_io(df, sep=sep, index=index, **kwargs)
    with lzma.open(path, "wb") as file:
        file.write(bytes_io.read())


def write_zipfile_csv(
    df: pandas.DataFrame,
    path: str | Path,
    inner_path: str,
    sep: str = "\t",
    index: bool = False,
    **kwargs: Any,
) -> None:
    """Write a dataframe to an inner CSV file to a zip archive.

    :param df: A dataframe
    :param path: The path to the resulting zip archive
    :param inner_path: The path inside the zip archive to write the dataframe
    :param sep: The separator in the dataframe. Overrides Pandas default to use a tab.
    :param index: Should the index be output? Overrides the Pandas default to be false.
    :param kwargs: Additional kwargs to pass to :func:`get_df_io` and transitively to
        :func:`pandas.DataFrame.to_csv`.
    """
    bytes_io = get_df_io(df, sep=sep, index=index, **kwargs)
    with open_zipfile(path, inner_path, operation="write", representation="binary") as file:
        file.write(bytes_io.read())


def read_zipfile_csv(
    path: str | Path, inner_path: str, sep: str = "\t", **kwargs: Any
) -> pandas.DataFrame:
    """Read an inner CSV file from a zip archive.

    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the dataframe
    :param sep: The separator in the dataframe. Overrides Pandas default to use a tab.
    :param kwargs: Additional kwargs to pass to :func:`pandas.read_csv`.

    :returns: A dataframe
    """
    import pandas as pd

    with open_zipfile(path, inner_path, representation="text", operation="read") as file:
        return pd.read_csv(file, sep=sep, **kwargs)


# docstr-coverage:excused `overload`
@typing.overload
@contextlib.contextmanager
def open_zipfile(
    path: str | Path,
    inner_path: str,
    *,
    operation: Operation = ...,
    representation: Literal["text"],
    zipfile_kwargs: Mapping[str, Any] | None = ...,
    open_kwargs: Mapping[str, Any] | None = ...,
) -> Generator[typing.TextIO, None, None]: ...


# docstr-coverage:excused `overload`
@typing.overload
@contextlib.contextmanager
def open_zipfile(
    path: str | Path,
    inner_path: str,
    *,
    operation: Operation = ...,
    representation: Literal["binary"],
    zipfile_kwargs: Mapping[str, Any] | None = ...,
    open_kwargs: Mapping[str, Any] | None = ...,
) -> Generator[typing.BinaryIO, None, None]: ...


@contextlib.contextmanager
def open_zipfile(
    path: str | Path,
    inner_path: str,
    *,
    operation: Operation = "read",
    representation: Representation,
    zipfile_kwargs: Mapping[str, Any] | None = None,
    open_kwargs: Mapping[str, Any] | None = None,
) -> Generator[typing.TextIO, None, None] | Generator[typing.BinaryIO, None, None]:
    """Open a zipfile."""
    mode: Literal["r", "w"]
    if operation == "read":
        mode = "r"
    elif operation == "write":
        mode = "w"
    else:
        raise InvalidOperationError(operation)

    # there might be a better way to deal with the mode here
    with zipfile.ZipFile(file=path, mode=mode, **(zipfile_kwargs or {})) as zip_file:
        with zip_file.open(inner_path, mode=mode, **(open_kwargs or {})) as binary_file:
            if representation == "text":
                with io.TextIOWrapper(binary_file, encoding="utf-8") as text_file:
                    yield text_file
            elif representation == "binary":
                yield cast(typing.BinaryIO, binary_file)
            else:
                raise InvalidRepresentationError(representation)


@contextlib.contextmanager
def open_tarfile(
    path: str | Path,
    inner_path: str,
    *,
    operation: Operation = "read",
    representation: Representation = "binary",
    open_kwargs: Mapping[str, Any] | None = None,
) -> Generator[typing.IO[bytes], None, None]:
    """Open a tar file."""
    if representation != "binary":
        raise NotImplementedError("tarfile must use binary representation")

    if operation == "read":
        with tarfile.open(path, "r", **(open_kwargs or {})) as tar:
            member = tar.getmember(inner_path)
            file = tar.extractfile(member)
            if file is None:
                raise FileNotFoundError(f"could not find {inner_path} in tarfile {path}")
            yield file
    elif operation == "write":
        file = BytesIO()
        yield file
        file.seek(0)
        tarinfo = tarfile.TarInfo(name=inner_path)
        tarinfo.size = len(file.getbuffer())
        with tarfile.TarFile(path, mode="w") as tar_file:
            tar_file.addfile(tarinfo, file)
    else:
        raise InvalidOperationError(operation)


class InvalidRepresentationError(ValueError):
    """Raised when passing an invalid representation."""

    def __init__(self, representation: str) -> None:
        """Instantiate the exception."""
        self.representation = representation

    def __str__(self) -> str:
        """Create a string for the exception."""
        return (
            f"Invalid representation: {self.representation}. "
            f"Should be one of {REPRESENTATION_VALUES}."
        )


class InvalidOperationError(ValueError):
    """Raised when passing an invalid operation."""

    def __init__(self, operation: str) -> None:
        """Instantiate the exception."""
        self.operation = operation

    def __str__(self) -> str:
        """Create a string for the exception."""
        return f"Invalid operation: {self.operation}. Should be one of {OPERATION_VALUES}."


@contextlib.contextmanager
def open_zip_reader(
    path: str | Path, inner_path: str, delimiter: str = "\t", **kwargs: Any
) -> Generator[Reader, None, None]:
    """Read an inner CSV file from a zip archive.

    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the CSV
    :param delimiter: The separator in the CSV. Defaults to tab.
    :param kwargs: Additional kwargs to pass to :func:`csv.reader`.

    :returns: A reader over the file
    """
    with open_zipfile(path, inner_path, representation="text") as file:
        yield csv.reader(file, delimiter=delimiter, **kwargs)


@contextlib.contextmanager
def open_zip_writer(
    path: str | Path, inner_path: str, delimiter: str = "\t", **kwargs: Any
) -> Generator[Writer, None, None]:
    """Open a writer for an inner CSV file from a zip archive.

    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the CSV
    :param delimiter: The separator in the CSV. Defaults to tab.
    :param kwargs: Additional kwargs to pass to :func:`csv.writer`.

    :returns: A writer over the file
    """
    with open_zipfile(path, inner_path, operation="write", representation="text") as file:
        yield csv.writer(file, delimiter=delimiter, **kwargs)


def write_zipfile_xml(
    element_tree: lxml.etree.ElementTree,
    path: str | Path,
    inner_path: str,
    **kwargs: Any,
) -> None:
    """Write an XML element tree to an inner XML file to a zip archive.

    :param element_tree: An XML element tree
    :param path: The path to the resulting zip archive
    :param inner_path: The path inside the zip archive to write the XML element
    :param kwargs: Additional kwargs to pass to :func:`lxml.etree.tostring`
    """
    from lxml import etree

    kwargs.setdefault("pretty_print", True)
    with open_zipfile(path, inner_path, operation="write", representation="binary") as file:
        file.write(etree.tostring(element_tree, **kwargs))


def read_zipfile_xml(path: str | Path, inner_path: str, **kwargs: Any) -> lxml.etree.ElementTree:
    """Read an inner XML file from a zip archive.

    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the xml file
    :param kwargs: Additional kwargs to pass to :func:`lxml.etree.parse`

    :returns: An XML element tree
    """
    from lxml import etree

    with open_zipfile(path, inner_path, operation="read", representation="binary") as file:
        return etree.parse(file, **kwargs)


def write_zipfile_np(
    arr: numpy.typing.ArrayLike,
    path: str | Path,
    inner_path: str,
    **kwargs: Any,
) -> None:
    """Write a dataframe to an inner CSV file to a zip archive.

    :param arr: Array-like
    :param path: The path to the resulting zip archive
    :param inner_path: The path inside the zip archive to write the dataframe
    :param kwargs: Additional kwargs to pass to :func:`get_np_io` and transitively to
        :func:`numpy.save`.
    """
    import numpy as np

    with open_zipfile(path, inner_path, operation="write", representation="binary") as file:
        np.save(file, arr, **kwargs)


def read_zip_np(path: str | Path, inner_path: str, **kwargs: Any) -> numpy.typing.ArrayLike:
    """Read an inner numpy array-like from a zip archive.

    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the dataframe
    :param kwargs: Additional kwargs to pass to :func:`numpy.load`.

    :returns: A numpy array or other object
    """
    import numpy as np

    with open_zipfile(path, inner_path, operation="read", representation="binary") as file:
        return cast(np.typing.ArrayLike, np.load(file, **kwargs))


def read_zipfile_rdf(path: str | Path, inner_path: str, **kwargs: Any) -> rdflib.Graph:
    """Read an inner RDF file from a zip archive.

    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the dataframe
    :param kwargs: Additional kwargs to pass to :meth:`rdflib.Graph.parse`.

    :returns: A graph
    """
    import rdflib

    graph = rdflib.Graph()
    with open_zipfile(path, inner_path, operation="read", representation="binary") as file:
        graph.parse(file, **kwargs)
    return graph


def write_zipfile_rdf(
    graph: rdflib.Graph, path: str | Path, inner_path: str, **kwargs: Any
) -> None:
    """Read an inner RDF file from a zip archive.

    :param graph: The graph to write
    :param path: The path to the zip archive
    :param inner_path: The path inside the zip archive to the dataframe
    :param kwargs: Additional kwargs to pass to :meth:`rdflib.Graph.parse`.
    """
    with open_zipfile(path, inner_path, operation="write", representation="binary") as file:
        graph.serialize(file, **kwargs)


def write_tarfile_csv(
    df: pandas.DataFrame,
    path: str | Path,
    inner_path: str,
    sep: str = "\t",
    index: bool = False,
    **kwargs: Any,
) -> None:
    """Write a dataframe to an inner CSV file from a tar archive.

    :param df: A dataframe
    :param path: The path to the resulting tar archive
    :param inner_path: The path inside the tar archive to write the dataframe
    :param sep: The separator in the dataframe. Overrides Pandas default to use a tab.
    :param index: Should the index be output? Overrides the Pandas default to be false.
    :param kwargs: Additional kwargs to pass to :func:`get_df_io` and transitively to
        :func:`pandas.DataFrame.to_csv`.
    """
    with open_tarfile(path, inner_path, operation="write") as file:
        df.to_csv(file, sep=sep, index=index, **kwargs)


def write_tarfile_xml(
    element_tree: lxml.etree.ElementTree,
    path: str | Path,
    inner_path: str,
    **kwargs: Any,
) -> None:
    """Write an XML document a tar archive.

    :param element_tree: An element
    :param path: The path to the resulting tar archive
    :param inner_path: The path inside the tar archive to write the dataframe
    :param kwargs: Additional kwargs to pass to :func:`lxml.etree.tostring`
    """
    from lxml import etree

    kwargs.setdefault("pretty_print", True)

    with open_tarfile(path, inner_path, operation="write") as file:
        file.write(etree.tostring(element_tree, **kwargs))


def read_tarfile_csv(
    path: str | Path, inner_path: str, sep: str = "\t", **kwargs: Any
) -> pandas.DataFrame:
    """Read an inner CSV file from a tar archive.

    :param path: The path to the tar archive
    :param inner_path: The path inside the tar archive to the dataframe
    :param sep: The separator in the dataframe. Overrides Pandas default to use a tab.
    :param kwargs: Additional kwargs to pass to :func:`pandas.read_csv`.

    :returns: A dataframe
    """
    import pandas as pd

    with open_tarfile(path, inner_path) as file:
        return pd.read_csv(file, sep=sep, **kwargs)


def read_tarfile_xml(path: str | Path, inner_path: str, **kwargs: Any) -> lxml.etree.ElementTree:
    """Read an inner XML file from a tar archive.

    :param path: The path to the tar archive
    :param inner_path: The path inside the tar archive to the xml file
    :param kwargs: Additional kwargs to pass to :func:`lxml.etree.parse`

    :returns: An XML element tree
    """
    from lxml import etree

    with open_tarfile(path, inner_path) as file:
        return etree.parse(file, **kwargs)


def read_rdf(path: str | Path, **kwargs: Any) -> rdflib.Graph:
    """Read an RDF file with :mod:`rdflib`.

    :param path: The path to the RDF file
    :param kwargs: Additional kwargs to pass to :func:`rdflib.Graph.parse`

    :returns: A parsed RDF graph
    """
    import rdflib

    graph = rdflib.Graph()
    with safe_open(path, representation="binary", operation="read") as file:
        graph.parse(file, **kwargs)
    return graph


def write_sql(df: pandas.DataFrame, name: str, path: str | Path, **kwargs: Any) -> None:
    """Write a dataframe as a SQL table.

    :param df: A dataframe
    :param name: The table the database to write to
    :param path: The path to the resulting tar archive
    :param kwargs: Additional keyword arguments to pass to
        :meth:`pandas.DataFrame.to_sql`
    """
    import sqlite3

    with contextlib.closing(sqlite3.connect(path)) as conn:
        df.to_sql(name, conn, **kwargs)


def get_commit(org: str, repo: str, provider: str = "git") -> str:
    """Get last commit hash for the given repo.

    :param org: The GitHub organization or owner
    :param repo: The GitHub repository name
    :param provider: The method for getting the most recent commit

    :returns: A commit hash's hex digest as a string

    :raises ValueError: if an invalid provider is given
    """
    if provider == "git":
        output = check_output(["git", "ls-remote", f"https://github.com/{org}/{repo}"])  # noqa
        lines = (line.strip().split("\t") for line in output.decode("utf8").splitlines())
        rv = next(line[0] for line in lines if line[1] == "HEAD")
    elif provider == "github":
        res = requests.get(f"https://api.github.com/repos/{org}/{repo}/branches/master", timeout=15)
        res_json = res.json()
        rv = res_json["commit"]["sha"]
    else:
        raise ValueError(f"invalid implementation: {provider}")
    return rv


CHUNK_SIZE = 32768
DOWNLOAD_URL = "https://docs.google.com/uc?export=download"
TOKEN_KEY = "download_warning"  # noqa:S105


def download_from_google(
    file_id: str,
    path: str | Path,
    force: bool = True,
    clean_on_failure: bool = True,
    hexdigests: Mapping[str, str] | None = None,
) -> None:
    """Download a file from google drive.

    Implementation inspired by https://github.com/ndrplz/google-drive-downloader.

    :param file_id: The google file identifier
    :param path: The place to write the file
    :param force: If false and the file already exists, will not re-download.
    :param clean_on_failure: If true, will delete the file on any exception raised
        during download
    :param hexdigests: The expected hexdigests as (algorithm_name, expected_hex_digest)
        pairs.

    :raises Exception: Thrown if an error besides a keyboard interrupt is thrown during
        download
    :raises KeyboardInterrupt: If a keyboard interrupt is thrown during download
    :raises UnexpectedDirectory: If a directory is given for the ``path`` argument
    """
    path = Path(path).resolve()

    if path.is_dir():
        raise UnexpectedDirectoryError(path)
    if path.is_file() and not force:
        raise_on_digest_mismatch(path=path, hexdigests=hexdigests)
        logger.debug("did not re-download %s from Google ID %s", path, file_id)
        return

    try:
        with requests.Session() as sess:
            res = sess.get(DOWNLOAD_URL, params={"id": file_id}, stream=True)
            token = _get_confirm_token(res)
            res = sess.get(DOWNLOAD_URL, params={"id": file_id, "confirm": token}, stream=True)
            with path.open("wb") as file:
                for chunk in tqdm(res.iter_content(CHUNK_SIZE), desc="writing", unit="chunk"):
                    if chunk:  # filter out keep-alive new chunks
                        file.write(chunk)
    except (Exception, KeyboardInterrupt):
        if clean_on_failure:
            _unlink(path)
        raise

    raise_on_digest_mismatch(path=path, hexdigests=hexdigests)


def _get_confirm_token(res: requests.Response) -> str:
    for key, value in res.cookies.items():
        if key.startswith(TOKEN_KEY):
            return value
    raise ValueError(f"no token found with key {TOKEN_KEY} in cookies: {res.cookies}")


def download_from_s3(
    s3_bucket: str,
    s3_key: str,
    path: str | Path,
    client: None | botocore.client.BaseClient = None,
    client_kwargs: Mapping[str, Any] | None = None,
    download_file_kwargs: Mapping[str, Any] | None = None,
    force: bool = True,
    clean_on_failure: bool = True,
) -> None:
    """Download a file from S3.

    :param s3_bucket: The key inside the S3 bucket name
    :param s3_key: The key inside the S3 bucket
    :param path: The place to write the file
    :param client: A botocore client. If none given, one will be created automatically
    :param client_kwargs: Keyword arguments to be passed to the client on instantiation.
    :param download_file_kwargs: Keyword arguments to be passed to
        :func:`boto3.s3.transfer.S3Transfer.download_file`
    :param force: If false and the file already exists, will not re-download.
    :param clean_on_failure: If true, will delete the file on any exception raised
        during download

    :raises Exception: Thrown if an error besides a keyboard interrupt is thrown during
        download
    :raises KeyboardInterrupt: If a keyboard interrupt is thrown during download
    :raises UnexpectedDirectory: If a directory is given for the ``path`` argument
    """
    path = Path(path).resolve()

    if path.is_dir():
        raise UnexpectedDirectoryError(path)
    if path.is_file() and not force:
        logger.debug("did not re-download %s from %s %s", path, s3_bucket, s3_key)
        return

    try:
        import boto3.s3.transfer

        if client is None:
            import boto3
            import botocore.client

            client_kwargs = {} if client_kwargs is None else dict(client_kwargs)
            client_kwargs.setdefault(
                "config", botocore.client.Config(signature_version=botocore.UNSIGNED)
            )
            client = boto3.client("s3", **client_kwargs)

        download_file_kwargs = {} if download_file_kwargs is None else dict(download_file_kwargs)
        download_file_kwargs.setdefault(
            "Config", boto3.s3.transfer.TransferConfig(use_threads=False)
        )
        client.download_file(s3_bucket, s3_key, path.as_posix(), **download_file_kwargs)
    except (Exception, KeyboardInterrupt):
        if clean_on_failure:
            _unlink(path)
        raise


def _unlink(path: str | Path) -> None:
    # python 3.6 does not have pathlib.Path.unlink, smh
    try:
        os.remove(path)
    except OSError:
        pass  # if the file can't be deleted then no problem


def get_name() -> str:
    """Get the PyStow home directory name.

    :returns: The name of the pystow home directory, either loaded from the
        :data:`PYSTOW_NAME_ENVVAR`` environment variable or given by the default value
        :data:`PYSTOW_NAME_DEFAULT`.
    """
    return os.getenv(PYSTOW_NAME_ENVVAR, default=PYSTOW_NAME_DEFAULT)


def use_appdirs() -> bool:
    """Check if X Desktop Group (XDG) compatibility is requested.

    :returns: If the :data:`PYSTOW_USE_APPDIRS` is set to ``true`` in the environment.
    """
    return os.getenv(PYSTOW_USE_APPDIRS) in {"true", "True"}


def get_home(ensure_exists: bool = True) -> Path:
    """Get the PyStow home directory.

    :param ensure_exists: If true, ensures the directory is created

    :returns: A path object representing the pystow home directory, as one of:

        1. :data:`PYSTOW_HOME_ENVVAR` environment variable or
        2. The user data directory defined by :mod:`appdirs` if the
           :data:`PYSTOW_USE_APPDIRS` environment variable is set to ``true`` or
        3. The default directory constructed in the user's home directory plus what's
           returned by :func:`get_name`.
    """
    if use_appdirs():
        from appdirs import user_data_dir

        default = Path(user_data_dir())
    else:
        default = Path.home() / get_name()
    return getenv_path(PYSTOW_HOME_ENVVAR, default, ensure_exists=ensure_exists)


def get_base(key: str, ensure_exists: bool = True) -> Path:
    """Get the base directory for a module.

    :param key: The name of the module. No funny characters. The envvar <key>_HOME where
        key is uppercased is checked first before using the default home directory.
    :param ensure_exists: Should all directories be created automatically? Defaults to
        true.

    :returns: The path to the given

    :raises ValueError: if the key is invalid (e.g., has a dot in it)
    """
    if "." in key:
        raise ValueError(f"The module should not have a dot in it: {key}")
    envvar = f"{key.upper()}_HOME"
    if use_appdirs():
        from appdirs import user_data_dir

        default = Path(user_data_dir(appname=key))
    else:
        default = get_home(ensure_exists=False) / key
    return getenv_path(envvar, default, ensure_exists=ensure_exists)


def ensure_readme() -> None:
    """Ensure there's a README in the PyStow data directory.

    :raises PermissionError: If the script calling this function does not have adequate
        permissions to write a file into the PyStow home directory.
    """
    try:
        readme_path = get_home(ensure_exists=True).joinpath("README.md")
    except PermissionError as e:
        raise PermissionError(
            "PyStow was not able to create its home directory in due to a lack of "
            "permissions. This can happen, e.g., if you're working on a server where you don't "
            "have full rights. See https://pystow.readthedocs.io/en/latest/installation.html#"
            "configuration for instructions on choosing a different home folder location for "
            "PyStow to somewhere where you have write permissions."
        ) from e
    if readme_path.is_file():
        return
    with readme_path.open("w", encoding="utf8") as file:
        print(README_TEXT, file=file)


def path_to_sqlite(path: str | Path) -> str:
    """Convert a path to a SQLite connection string.

    :param path: A path to a SQLite database file

    :returns: A standard connection string to the database
    """
    path = Path(path).expanduser().resolve()
    return f"sqlite:///{path.as_posix()}"


def gunzip(source: str | Path, target: str | Path) -> None:
    """Unzip a file in the source to the target.

    :param source: The path to an input file
    :param target: The path to an output file
    """
    with gzip.open(source, "rb") as in_file, open(target, "wb") as out_file:
        shutil.copyfileobj(in_file, out_file)


# docstr-coverage:excused `overload`
@typing.overload
@contextlib.contextmanager
def safe_open(
    path: str | Path, *, operation: Operation = ..., representation: Literal["text"] = "text"
) -> Generator[typing.TextIO, None, None]: ...


# docstr-coverage:excused `overload`
@typing.overload
@contextlib.contextmanager
def safe_open(
    path: str | Path, *, operation: Operation = ..., representation: Literal["binary"] = "binary"
) -> Generator[typing.BinaryIO, None, None]: ...


@contextlib.contextmanager
def safe_open(
    path: str | Path, *, operation: Operation = "read", representation: Representation = "text"
) -> Generator[typing.TextIO, None, None] | Generator[typing.BinaryIO, None, None]:
    """Safely open a file for reading or writing text."""
    if operation not in OPERATION_VALUES:
        raise InvalidOperationError(operation)
    if representation not in REPRESENTATION_VALUES:
        raise InvalidRepresentationError(representation)

    mode = MODE_MAP[operation, representation]
    path = Path(path).expanduser().resolve()
    if path.suffix.endswith(".gz"):
        with gzip.open(path, mode=mode) as file:
            yield file  # type:ignore
    else:
        with open(path, mode=mode) as file:
            yield file  # type:ignore


@contextlib.contextmanager
def safe_open_writer(
    f: str | Path | TextIO, *, delimiter: str = "\t", **kwargs: Any
) -> Generator[Writer, None, None]:
    """Open a CSV writer, wrapping :func:`csv.writer`.

    :param f: A path to a file, or an already open text-based IO object
    :param delimiter: The delimiter for writing to CSV
    :param kwargs: Keyword arguments to pass to :func:`csv.writer`

    :yields: A CSV writer object, constructed from :func:`csv.writer`
    """
    if isinstance(f, (str, Path)):
        with safe_open(f, operation="write", representation="text") as file:
            yield csv.writer(file, delimiter=delimiter, **kwargs)
    else:
        yield csv.writer(f, delimiter=delimiter, **kwargs)


@contextlib.contextmanager
def safe_open_dict_writer(
    f: str | Path | TextIO,
    fieldnames: typing.Sequence[str],
    *,
    delimiter: str = "\t",
    **kwargs: Any,
) -> Generator[csv.DictWriter[str], None, None]:
    """Open a CSV dictionary writer, wrapping :func:`csv.DictWriter`.

    :param f: A path to a file, or an already open text-based IO object
    :param fieldnames: A path to a file, or an already open text-based IO object
    :param delimiter: The delimiter for writing to CSV
    :param kwargs: Keyword arguments to pass to :func:`csv.DictWriter`

    :yields: A CSV dictionary writer object, constructed from :func:`csv.DictWriter`
    """
    if isinstance(f, (str, Path)):
        with safe_open(f, operation="write", representation="text") as file:
            yield csv.DictWriter(file, fieldnames, delimiter=delimiter, **kwargs)
    else:
        yield csv.DictWriter(f, fieldnames, delimiter=delimiter, **kwargs)


@contextlib.contextmanager
def safe_open_reader(
    f: str | Path | TextIO, *, delimiter: str = "\t", **kwargs: Any
) -> Generator[Reader, None, None]:
    """Open a CSV reader, wrapping :func:`csv.reader`.

    :param f: A path to a file, or an already open text-based IO object
    :param delimiter: The delimiter for writing to CSV
    :param kwargs: Keyword arguments to pass to :func:`csv.reader`

    :yields: A CSV reader object, constructed from :func:`csv.reader`
    """
    if isinstance(f, (str, Path)):
        with safe_open(f, operation="read", representation="text") as file:
            yield csv.reader(file, delimiter=delimiter, **kwargs)
    else:
        yield csv.reader(f, delimiter=delimiter, **kwargs)


@contextlib.contextmanager
def safe_open_dict_reader(
    f: str | Path | TextIO, *, delimiter: str = "\t", **kwargs: Any
) -> Generator[csv.DictReader[str], None, None]:
    """Open a CSV dictionary reader, wrapping :func:`csv.DictReader`.

    :param f: A path to a file, or an already open text-based IO object
    :param delimiter: The delimiter for writing to CSV
    :param kwargs: Keyword arguments to pass to :func:`csv.DictReader`

    :yields: A CSV reader object, constructed from :func:`csv.DictReader`
    """
    if isinstance(f, (str, Path)):
        with safe_open(f, operation="read", representation="text") as file:
            yield csv.DictReader(file, delimiter=delimiter, **kwargs)
    else:
        yield csv.DictReader(f, delimiter=delimiter, **kwargs)


def get_soup(
    url: str,
    *,
    verify: bool = True,
    timeout: TimeoutHint | None = None,
    user_agent: str | None = None,
) -> bs4.BeautifulSoup:
    """Get a beautiful soup parsed version of the given web page.

    :param url: The URL to download and parse with BeautifulSoup
    :param verify: Should SSL be used? This is almost always true, except for Ensembl,
        which makes a big pain
    :param timeout: How many integer seconds to wait for a response? Defaults to 15 if
        none given.
    :param user_agent: A custom user-agent to set, e.g., to avoid anti-crawling
        mechanisms

    :returns: A BeautifulSoup object
    """
    from bs4 import BeautifulSoup

    headers = {}
    if user_agent:
        headers["User-Agent"] = user_agent
    res = requests.get(url, verify=verify, timeout=timeout or 15, headers=headers)
    soup = BeautifulSoup(res.text, features="html.parser")
    return soup



"""PyStow: Easily pick a place to store data for your python package."""

from .api import (
    dump_df,
    dump_json,
    dump_pickle,
    dump_rdf,
    dump_xml,
    ensure,
    ensure_csv,
    ensure_custom,
    ensure_excel,
    ensure_from_google,
    ensure_from_s3,
    ensure_gunzip,
    ensure_json,
    ensure_json_bz2,
    ensure_nltk,
    ensure_open,
    ensure_open_bz2,
    ensure_open_gz,
    ensure_open_lzma,
    ensure_open_sqlite,
    ensure_open_sqlite_gz,
    ensure_open_tarfile,
    ensure_open_zip,
    ensure_pickle,
    ensure_pickle_gz,
    ensure_rdf,
    ensure_soup,
    ensure_tar_df,
    ensure_tar_xml,
    ensure_untar,
    ensure_xml,
    ensure_yaml,
    ensure_zip_df,
    ensure_zip_np,
    join,
    joinpath_sqlite,
    load_df,
    load_json,
    load_pickle,
    load_pickle_gz,
    load_rdf,
    load_xml,
    load_yaml,
    module,
    open,
    open_gz,
)
from .config_api import ConfigError, get_config, write_config
from .impl import Module, VersionHint
from .utils import ensure_readme

__all__ = [
    "ConfigError",
    "Module",
    "VersionHint",
    "dump_df",
    "dump_json",
    "dump_pickle",
    "dump_rdf",
    "dump_xml",
    "ensure",
    "ensure_csv",
    "ensure_custom",
    "ensure_excel",
    "ensure_from_google",
    "ensure_from_s3",
    "ensure_gunzip",
    "ensure_json",
    "ensure_json_bz2",
    "ensure_nltk",
    "ensure_open",
    "ensure_open_bz2",
    "ensure_open_gz",
    "ensure_open_lzma",
    "ensure_open_sqlite",
    "ensure_open_sqlite_gz",
    "ensure_open_tarfile",
    "ensure_open_zip",
    "ensure_pickle",
    "ensure_pickle_gz",
    "ensure_rdf",
    "ensure_soup",
    "ensure_tar_df",
    "ensure_tar_xml",
    "ensure_untar",
    "ensure_xml",
    "ensure_yaml",
    "ensure_zip_df",
    "ensure_zip_np",
    "get_config",
    "join",
    "joinpath_sqlite",
    "load_df",
    "load_json",
    "load_pickle",
    "load_pickle_gz",
    "load_rdf",
    "load_xml",
    "load_yaml",
    "module",
    "open",
    "open_gz",
    "write_config",
]

ensure_readme()

del ensure_readme



"""Utilities for the web app."""

from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path
    from subprocess import CompletedProcess

__all__ = [
    "clone_github_tempdir",
    "clone_tempdir",
    "commit",
    "create_branch",
    "fetch",
    "get_current_branch",
    "git",
    "has_local_branch",
    "is_likely_default_branch",
    "push",
]


@contextmanager
def clone_github_tempdir(owner: str, repo: str) -> Generator[Path, None, None]:
    """Temporarily clone a repository from a URL."""
    url = f"https://github.com/{owner}/{repo}.git"
    with clone_tempdir(url) as directory:
        yield directory


@contextmanager
def clone_tempdir(url: str) -> Generator[Path, None, None]:
    """Temporarily clone a repository from a URL."""
    import shutil
    import tempfile
    from pathlib import Path

    with tempfile.TemporaryDirectory() as directory_:
        directory = Path(directory_)
        clone(directory, url)
        yield directory
        shutil.rmtree(directory)


GitReturn: TypeAlias = "CompletedProcess[str]"


def git(directory: Path, *args: str) -> GitReturn:
    """Run the git command with the given arguments in the given directory."""
    return _check_output("git", *args, directory=directory)


def clone(directory: Path, url: str) -> GitReturn:
    """Clone the git repository with the given URL."""
    args = ["git", "clone", url, directory.as_posix()]
    return _check_output(*args, directory=directory)


def _check_output(*args: str, directory: Path | None = None) -> GitReturn:
    from subprocess import run

    return run(  # noqa: S603
        args,
        cwd=None if directory is None else directory.as_posix(),
        capture_output=True,
        text=True,
    )


def commit(directory: Path, message: str) -> GitReturn:
    """Make a commit with the following message."""
    return git(directory, "commit", "-m", message, "-a")


def push(directory: Path, branch: str | None = None) -> GitReturn:
    """Push the git repo."""
    if branch is not None:
        return git(directory, "push", "origin", branch)
    else:
        return git(directory, "push")


def fetch(directory: Path) -> GitReturn:
    """Fetch the git repo."""
    return git(directory, "fetch", "--all")


def has_local_branch(directory: Path, branch: str) -> bool:
    """Check if a branch exists in the git repo."""
    completed_process = git(directory, "show-ref", "--verify", "--quiet", f"refs/heads/{branch}")
    return completed_process.returncode == 0


def get_current_branch(directory: Path) -> str:
    """Return if on the master/main branch."""
    completed_process = git(directory, "rev-parse", "--abbrev-ref", "HEAD")
    return completed_process.stdout.strip()


def create_branch(directory: Path, branch: str) -> GitReturn:
    """Create a branch and check it out."""
    return git(directory, "checkout", "-b", branch)


LIKELY_DEFAULT_BRANCHES = {"master", "main"}


def is_likely_default_branch(directory: Path) -> bool:
    """Guess if the current branch is the "default"."""
    return get_current_branch(directory) in LIKELY_DEFAULT_BRANCHES



"""Utilities for caching files."""

from __future__ import annotations

import functools
import json
import logging
import pickle
from abc import ABC, abstractmethod
from collections.abc import Callable, MutableMapping
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    cast,
)

if TYPE_CHECKING:
    import pandas as pd

__all__ = [
    # Classes
    "Cached",
    "CachedCollection",
    "CachedDataFrame",
    "CachedJSON",
    "CachedPickle",
    # Types
    "Getter",
]

logger = logging.getLogger(__name__)

JSONType = dict[str, Any] | list[Any]
X = TypeVar("X")
Getter = Callable[[], X]


class Cached(Generic[X], ABC):
    """Caching decorator."""

    def __init__(
        self,
        path: str | Path,
        *,
        force: bool = False,
        cache: bool = True,
    ) -> None:
        """Instantiate the decorator.

        :param path: The path to the cache for the file
        :param cache: Should caching be done? Defaults to true, turn off for debugging
            purposes
        :param force: Should a pre-existing file be disregared/overwritten?
        """
        self.path = Path(path)
        self.force = force
        self.cache = cache

    def __call__(self, func: Getter[X]) -> Getter[X]:
        """Apply this instance as a decorator.

        :param func: The function to wrap

        :returns: A wrapped function
        """

        @functools.wraps(func)
        def _wrapped() -> X:
            if not self.cache:
                return func()

            if self.path.is_file() and not self.force:
                return self.load()
            logger.debug("no cache found at %s", self.path)
            rv = func()
            logger.debug("writing cache to %s", self.path)
            self.dump(rv)
            return rv

        return _wrapped

    @abstractmethod
    def load(self) -> X:
        """Load data from the cache (typically by opening a file at the given path)."""

    @abstractmethod
    def dump(self, rv: X) -> None:
        """Dump data to the cache (typically by opening a file at the given path).

        :param rv: The data to dump
        """


class CachedJSON(Cached[JSONType]):
    """Make a function lazily cache its return value as JSON."""

    def load(self) -> JSONType:
        """Load data from the cache as JSON.

        :returns: A python object with JSON-like data from the cache
        """
        with open(self.path) as file:
            return cast(JSONType, json.load(file))

    def dump(self, rv: JSONType) -> None:
        """Dump data to the cache as JSON.

        :param rv: The JSON data to dump
        """
        with open(self.path, "w") as file:
            json.dump(rv, file, indent=2)


class CachedPickle(Cached[Any]):
    """Make a function lazily cache its return value as a pickle."""

    def load(self) -> Any:
        """Load data from the cache as a pickle.

        :returns: A python object loaded from the cache
        """
        with open(self.path, "rb") as file:
            return pickle.load(file)

    def dump(self, rv: Any) -> None:
        """Dump data to the cache as a pickle.

        :param rv: The arbitrary python object to dump
        """
        with open(self.path, "wb") as file:
            pickle.dump(rv, file, protocol=pickle.HIGHEST_PROTOCOL)


class CachedCollection(Cached[list[str]]):
    """Make a function lazily cache its return value as file."""

    def load(self) -> list[str]:
        """Load data from the cache as a list of strings.

        :returns: A list of strings loaded from the cache
        """
        with open(self.path) as file:
            return [line.strip() for line in file]

    def dump(self, rv: list[str]) -> None:
        """Dump data to the cache as a list of strings.

        :param rv: The list of strings to dump
        """
        with open(self.path, "w") as file:
            for line in rv:
                print(line, file=file)


class CachedDataFrame(Cached["pd.DataFrame"]):
    """Make a function lazily cache its return value as a dataframe."""

    def __init__(
        self,
        path: str | Path,
        cache: bool = True,
        force: bool = False,
        sep: str | None = None,
        dtype: Any | None = None,
        read_csv_kwargs: MutableMapping[str, Any] | None = None,
    ) -> None:
        """Instantiate the decorator.

        :param path: The path to the cache for the file
        :param force: Should a pre-existing file be disregared/overwritten?
        :param sep: The separator. Defaults to TSV, since this is the only reasonable
            default.
        :param dtype: A shortcut for setting the dtype
        :param read_csv_kwargs: Additional kwargs to pass to :func:`pd.read_csv`.

        :raises ValueError: if sep is given as a kwarg and also in ``read_csv_kwargs``.
        """
        super().__init__(path=path, cache=cache, force=force)
        self.read_csv_kwargs = read_csv_kwargs or {}
        if "sep" not in self.read_csv_kwargs:
            self.sep = sep or "\t"
        elif sep is not None:
            raise ValueError
        else:
            self.sep = self.read_csv_kwargs.pop("sep")
        if dtype is not None:
            if "dtype" in self.read_csv_kwargs:
                raise ValueError
            self.read_csv_kwargs["dtype"] = dtype
        self.read_csv_kwargs.setdefault("keep_default_na", False)

    def load(self) -> pd.DataFrame:
        """Load data from the cache as a dataframe.

        :returns: A dataframe loaded from the cache.
        """
        import pandas as pd

        return pd.read_csv(
            self.path,
            sep=self.sep,
            **self.read_csv_kwargs,
        )

    def dump(self, rv: pd.DataFrame) -> None:
        """Dump data to the cache as a dataframe.

        :param rv: The dataframe to dump
        """
        rv.to_csv(self.path, sep=self.sep, index=False)



"""Module implementation."""

from __future__ import annotations

import bz2
import gzip
import io
import json
import logging
import lzma
import os
import pickle
import sqlite3
import tarfile
import typing
from collections.abc import Callable, Generator, Mapping, Sequence
from contextlib import closing, contextmanager
from io import BytesIO, StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast, overload

from . import utils
from .constants import JSON, Provider
from .utils import (
    base_from_gzip_name,
    download_from_google,
    download_from_s3,
    get_base,
    get_mode_pair,
    gunzip,
    mkdir,
    name_from_s3_key,
    name_from_url,
    open_tarfile,
    open_zipfile,
    path_to_sqlite,
    read_rdf,
    read_tarfile_csv,
    read_tarfile_xml,
    read_zip_np,
    read_zipfile_csv,
)

if TYPE_CHECKING:
    import botocore.client
    import bs4
    import lxml.etree
    import numpy
    import pandas as pd
    import rdflib

__all__ = [
    "Module",
    "VersionHint",
]

logger = logging.getLogger(__name__)

#: A type hint for something that can be passed to the
#: `version` argument of Module.join, Module.ensure, etc.
VersionHint: TypeAlias = None | str | Callable[[], str | None]


class Module:
    """The class wrapping the directory lookup implementation."""

    def __init__(self, base: str | Path, ensure_exists: bool = True) -> None:
        """Initialize the module.

        :param base: The base directory for the module
        :param ensure_exists: Should the base directory be created automatically?
            Defaults to true.
        """
        self.base = Path(base)
        mkdir(self.base, ensure_exists=ensure_exists)

    @classmethod
    def from_key(cls, key: str, *subkeys: str, ensure_exists: bool = True) -> Module:
        """Get a module for the given directory or one of its subdirectories.

        :param key: The name of the module. No funny characters. The envvar <key>_HOME
            where key is uppercased is checked first before using the default home
            directory.
        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param ensure_exists: Should all directories be created automatically? Defaults
            to true.

        :returns: A module
        """
        base = get_base(key, ensure_exists=False)
        rv = cls(base=base, ensure_exists=ensure_exists)
        if subkeys:
            rv = rv.module(*subkeys, ensure_exists=ensure_exists)
        return rv

    def module(self, *subkeys: str, ensure_exists: bool = True) -> Module:
        """Get a module for a subdirectory of the current module.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param ensure_exists: Should all directories be created automatically? Defaults
            to true.

        :returns: A module representing the subdirectory based on the given ``subkeys``.
        """
        base = self.join(*subkeys, ensure_exists=False)
        return Module(base=base, ensure_exists=ensure_exists)

    def join(
        self,
        *subkeys: str,
        name: str | None = None,
        ensure_exists: bool = True,
        version: VersionHint = None,
    ) -> Path:
        """Get a subdirectory of the current module.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param ensure_exists: Should all directories be created automatically? Defaults
            to true.
        :param name: The name of the file (optional) inside the folder
        :param version: The optional version, or no-argument callable that returns an
            optional version. This is prepended before the subkeys.

            The following example describes how to store the versioned data from the
            Rhea database for biologically relevant chemical reactions.

            .. code-block::

                import pystow
                import requests

                def get_rhea_version() -> str:
                    res = requests.get("https://ftp.expasy.org/databases/rhea/rhea-release.properties")
                    _, _, version = res.text.splitlines()[0].partition("=")
                    return version

                # Assume you want to download the data from
                # ftp://ftp.expasy.org/databases/rhea/rdf/rhea.rdf.gz, make a path
                # with the same name
                module = pystow.module("rhea")
                path = module.join(name="rhea.rdf.gz", version=get_rhea_version)


        :returns: The path of the directory or subdirectory for the given module.
        """
        rv = self.base

        # if the version is given as a no-argument callable,
        # then it should be called and a version is returned
        if callable(version):
            version = version()
        if version:
            self._raise_for_invalid_version(version)
            subkeys = (version, *subkeys)

        if subkeys:
            rv = rv.joinpath(*subkeys)
            mkdir(rv, ensure_exists=ensure_exists)
        if name:
            rv = rv.joinpath(name)
        return rv

    @staticmethod
    def _raise_for_invalid_version(version: str) -> None:
        if "/" in version or os.sep in version:
            raise ValueError(
                f"slashes and `{os.sep}` not allowed in versions because of "
                f"conflicts with file path construction: {version}"
            )

    def joinpath_sqlite(self, *subkeys: str, name: str) -> str:
        """Get an SQLite database connection string.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the database file.

        :returns: A SQLite path string.
        """
        path = self.join(*subkeys, name=name, ensure_exists=True)
        return path_to_sqlite(path)

    def ensure(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        version: VersionHint = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
    ) -> Path:
        """Ensure a file is downloaded.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param version: The optional version, or no-argument callable that returns an
            optional version. This is prepended before the subkeys.

            The following example describes how to store the versioned data from the
            Rhea database for biologically relevant chemical reactions.

            .. code-block::

                import pystow
                import requests

                def get_rhea_version() -> str:
                    res = requests.get("https://ftp.expasy.org/databases/rhea/rhea-release.properties")
                    _, _, version = res.text.splitlines()[0].partition("=")
                    return version

                module = pystow.module("rhea")
                path = module.ensure(
                    url="ftp://ftp.expasy.org/databases/rhea/rdf/rhea.rdf.gz",
                    version=get_rhea_version,
                )

        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.

        :returns: The path of the file that has been downloaded (or already exists)
        """
        if name is None:
            name = name_from_url(url)
        path = self.join(*subkeys, name=name, version=version, ensure_exists=True)
        utils.download(
            url=url,
            path=path,
            force=force,
            **(download_kwargs or {}),
        )
        return path

    def ensure_soup(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        version: VersionHint = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["r", "rt", "w", "wt"] | Literal["rb", "wb"] = "r",
        open_kwargs: Mapping[str, Any] | None = None,
        beautiful_soup_kwargs: Mapping[str, Any] | None = None,
    ) -> bs4.BeautifulSoup:
        """Ensure a webpage is downloaded and parsed with :mod:`BeautifulSoup`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param version: The optional version, or no-argument callable that returns an
            optional version. This is prepended before the subkeys.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param beautiful_soup_kwargs: Additional keyword arguments passed to
            :class:`BeautifulSoup`

        :returns: An BeautifulSoup object

        .. note::

            If you don't need to cache, consider using :func:`pystow.utils.get_soup`
            instead.
        """
        from bs4 import BeautifulSoup

        beautiful_soup_kwargs = dict(beautiful_soup_kwargs or {})
        beautiful_soup_kwargs.setdefault("features", "html.parser")

        with self.ensure_open(
            *subkeys,
            url=url,
            name=name,
            version=version,
            force=force,
            download_kwargs=download_kwargs,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            soup = BeautifulSoup(file, **beautiful_soup_kwargs)
        return soup

    def ensure_custom(
        self,
        *subkeys: str,
        name: str,
        force: bool = False,
        provider: Provider,
        **kwargs: Any,
    ) -> Path:
        """Ensure a file is present, and run a custom create function otherwise.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The file name.
        :param force: Should the file be re-created, even if the path already exists?
        :param provider: The file provider. Will be run with the path as the first
            positional argument, if the file needs to be generated.
        :param kwargs: Additional keyword-based parameters passed to the provider.

        :returns: The path of the file that has been created (or already exists)

        :raises ValueError: If the provider was called but the file was not created by
            it.
        """
        path = self.join(*subkeys, name=name, ensure_exists=True)
        if path.is_file() and not force:
            return path
        provider(path, **kwargs)
        if not path.is_file():
            raise ValueError(f"Provider {provider} did not create the file at {path}!")
        return path

    def ensure_untar(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        directory: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        extract_kwargs: Mapping[str, Any] | None = None,
    ) -> Path:
        """Ensure a tar file is downloaded and unarchived.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param directory: Overrides the name of the directory into which the tar archive
            is extracted. If none given, will use the stem of the file name that gets
            downloaded.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param extract_kwargs: Keyword arguments to pass to
            :meth:`tarfile.TarFile.extract_all`.

        :returns: The path of the directory where the file that has been downloaded gets
            extracted to
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        if directory is None:
            # rhea-rxn.tar.gz -> rhea-rxn
            suffixes_len = sum(len(suffix) for suffix in path.suffixes)
            directory = path.name[:-suffixes_len]
        unzipped_path = path.parent.joinpath(directory)
        if unzipped_path.is_dir() and not force:
            return unzipped_path
        unzipped_path.mkdir(exist_ok=True, parents=True)
        with tarfile.open(path) as tar_file:
            tar_file.extractall(unzipped_path, **(extract_kwargs or {}))  # noqa:S202
        return unzipped_path

    def ensure_gunzip(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        autoclean: bool = True,
        download_kwargs: Mapping[str, Any] | None = None,
    ) -> Path:
        """Ensure a tar.gz file is downloaded and unarchived.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param autoclean: Should the zipped file be deleted?
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.

        :returns: The path of the directory where the file that has been downloaded gets
            extracted to
        """
        if name is None:
            name = name_from_url(url)
        gunzipped_name = base_from_gzip_name(name)
        gunzipped_path = self.join(*subkeys, name=gunzipped_name, ensure_exists=True)
        if gunzipped_path.is_file() and not force:
            return gunzipped_path
        path = self.ensure(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
        )
        gunzip(path, gunzipped_path)
        if autoclean:
            logger.info("removing original gzipped file %s", path)
            path.unlink()
        return gunzipped_path

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open(
        self,
        *subkeys: str,
        url: str,
        name: str | None,
        version: VersionHint = ...,
        force: bool,
        download_kwargs: Mapping[str, Any] | None,
        mode: Literal["r", "rt", "w", "wt"] = ...,
        open_kwargs: Mapping[str, Any] | None,
    ) -> Generator[StringIO, None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open(
        self,
        *subkeys: str,
        url: str,
        name: str | None,
        version: VersionHint = ...,
        force: bool,
        download_kwargs: Mapping[str, Any] | None,
        mode: Literal["rb", "wb"] = ...,
        open_kwargs: Mapping[str, Any] | None,
    ) -> Generator[BytesIO, None, None]: ...

    @contextmanager
    def ensure_open(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        version: VersionHint = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["r", "rt", "w", "wt"] | Literal["rb", "wb"] = "r",
        open_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[StringIO | BytesIO, None, None]:
        """Ensure a file is downloaded and open it.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param version: The optional version, or no-argument callable that returns an
            optional version. This is prepended before the subkeys.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys,
            url=url,
            name=name,
            version=version,
            force=force,
            download_kwargs=download_kwargs,
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with path.open(**open_kwargs) as file:
            yield file

    @staticmethod
    def _raise_for_mode_ensure_mismatch(
        mode: Literal["r", "rt", "w", "wt", "wb", "rb"], ensure_exists: bool
    ) -> None:
        """Raise an exception for a mismatch between the mode and the ensure status.

        :param mode: The file opening mode
        :param ensure_exists: The value for ensuring a file

        :raises ValueError: In the following situations:

            1. If the file should be opened in write mode, and it is not ensured to
               exist
            2. If the file should be opened in read mode, and it is ensured to exist.
               This is bad because it will create a file when there previously wasn't
               one
        """
        if "w" in mode and not ensure_exists:
            raise ValueError
        if "r" in mode and ensure_exists:
            raise ValueError

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def open(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["r", "rt", "w", "wt"] = ...,
        open_kwargs: Mapping[str, Any] | None = None,
        ensure_exists: bool,
    ) -> Generator[StringIO, None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def open(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["rb", "wb"] = ...,
        open_kwargs: Mapping[str, Any] | None = None,
        ensure_exists: bool,
    ) -> Generator[BytesIO, None, None]: ...

    @contextmanager
    def open(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["r", "rt", "w", "wt"] | Literal["rb", "wb"] = "r",
        open_kwargs: Mapping[str, Any] | None = None,
        ensure_exists: bool = False,
    ) -> Generator[StringIO | BytesIO, None, None]:
        """Open a file.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param ensure_exists: Should the directory the file is in be made? Set to true
            on write operations.

        :yields: An open file object.

        This function should be called inside a context manager like in the following

        .. code-block:: python

            import pystow

            with pystow.module("test").open(name="test.tsv", mode="w") as file:
                print("Test text!", file=file)
        """
        self._raise_for_mode_ensure_mismatch(mode, ensure_exists)
        path = self.join(*subkeys, name=name, ensure_exists=ensure_exists)
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with path.open(**open_kwargs) as file:
            yield file

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def open_gz(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["r", "w", "rt", "wt"] = ...,
        open_kwargs: Mapping[str, Any] | None,
        ensure_exists: bool,
    ) -> Generator[StringIO, None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def open_gz(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["rb", "wb"] = ...,
        open_kwargs: Mapping[str, Any] | None,
        ensure_exists: bool,
    ) -> Generator[BytesIO, None, None]: ...

    @contextmanager
    def open_gz(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["r", "w", "rt", "wt", "rb", "wb"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
        ensure_exists: bool = False,
    ) -> Generator[StringIO | BytesIO, None, None]:
        """Open a gzipped file that exists already.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`gzip.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
        :param ensure_exists: Should the file be made? Set to true on write operations.

        :yields: An open file object
        """
        self._raise_for_mode_ensure_mismatch(mode, ensure_exists)
        path = self.join(*subkeys, name=name, ensure_exists=ensure_exists)
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with gzip.open(path, **open_kwargs) as file:
            yield file

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_lzma(
        self,
        *subkeys: str,
        url: str,
        name: str | None,
        force: bool,
        download_kwargs: Mapping[str, Any] | None,
        mode: Literal["r", "w", "rt", "wt"] = "rt",
        open_kwargs: Mapping[str, Any] | None,
    ) -> Generator[io.TextIOWrapper[lzma.LZMAFile], None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_lzma(
        self,
        *subkeys: str,
        url: str,
        name: str | None,
        force: bool,
        download_kwargs: Mapping[str, Any] | None,
        mode: Literal["rb", "wb"] = ...,
        open_kwargs: Mapping[str, Any] | None,
    ) -> Generator[lzma.LZMAFile, None, None]: ...

    @contextmanager
    def ensure_open_lzma(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["r", "rb", "w", "wb", "rt", "wt"] = "rt",
        open_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[lzma.LZMAFile | io.TextIOWrapper[lzma.LZMAFile], None, None]:
        """Ensure a LZMA-compressed file is downloaded and open a file inside it.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`lzma.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`lzma.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with lzma.open(path, **open_kwargs) as file:
            yield file

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_tarfile(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = ...,
        force: bool = ...,
        download_kwargs: Mapping[str, Any] | None = ...,
        mode: Literal["rt"] = ...,
        open_kwargs: Mapping[str, Any] | None = ...,
    ) -> Generator[typing.TextIO, None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_tarfile(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = ...,
        force: bool = ...,
        download_kwargs: Mapping[str, Any] | None = ...,
        mode: Literal["r", "rb"] = ...,
        open_kwargs: Mapping[str, Any] | None = ...,
    ) -> Generator[typing.IO[bytes], None, None]: ...

    @contextmanager
    def ensure_open_tarfile(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["r", "rb", "rt"] = "r",
        open_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[typing.TextIO, None, None] | Generator[typing.IO[bytes], None, None]:
        """Ensure a tar file is downloaded and open a file inside it.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param inner_path: The relative path to the file inside the archive
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`tarfile.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`tarfile.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        operation, representation = get_mode_pair(open_kwargs.pop("mode"), interpretation="binary")

        with open_tarfile(
            path=path,
            inner_path=inner_path,
            operation=operation,
            representation=representation,
            open_kwargs=open_kwargs,
        ) as file:
            yield file

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_zip(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = ...,
        force: bool = ...,
        download_kwargs: Mapping[str, Any] | None = ...,
        mode: Literal["r", "rb"] = ...,
        open_kwargs: Mapping[str, Any] | None = ...,
    ) -> Generator[typing.BinaryIO, None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_zip(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = ...,
        force: bool = ...,
        download_kwargs: Mapping[str, Any] | None = ...,
        mode: Literal["rt"] = ...,
        open_kwargs: Mapping[str, Any] | None = ...,
    ) -> Generator[typing.TextIO, None, None]: ...

    @contextmanager
    def ensure_open_zip(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["r", "rb", "rt"] = "r",
        zipfile_kwargs: Mapping[str, Any] | None = None,
        open_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[typing.TextIO, None, None] | Generator[typing.BinaryIO, None, None]:
        """Ensure a file is downloaded then open it with :mod:`zipfile`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param inner_path: The relative path to the file inside the archive
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`zipfile.open`. Defaults to bytes mode
            for ``r`` and ``w``.
        :param zipfile_kwargs: Additional keyword arguments passed to :class:`zipfile.ZipFile`
        :param open_kwargs: Additional keyword arguments passed to :func:`zipfile.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )

        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        operation, representation = get_mode_pair(open_kwargs.pop("mode"), interpretation="binary")

        with open_zipfile(
            path=path,
            inner_path=inner_path,
            operation=operation,
            representation=representation,
            zipfile_kwargs=zipfile_kwargs,
            open_kwargs=open_kwargs,
        ) as file:
            yield file

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_gz(
        self,
        *subkeys: str,
        url: str,
        name: str | None,
        force: bool,
        download_kwargs: Mapping[str, Any] | None,
        mode: Literal["r", "w", "rt", "wt"] = ...,
        open_kwargs: Mapping[str, Any] | None,
    ) -> Generator[StringIO, None, None]: ...

    # docstr-coverage:excused `overload`
    @overload
    @contextmanager
    def ensure_open_gz(
        self,
        *subkeys: str,
        url: str,
        name: str | None,
        force: bool,
        download_kwargs: Mapping[str, Any] | None,
        mode: Literal[
            "rb",
            "wb",
        ] = ...,
        open_kwargs: Mapping[str, Any] | None,
    ) -> Generator[BytesIO, None, None]: ...

    @contextmanager
    def ensure_open_gz(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["r", "rb", "w", "wb", "rt", "wt"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[StringIO | BytesIO, None, None]:
        """Ensure a gzipped file is downloaded and open a file inside it.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`gzip.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with gzip.open(path, **open_kwargs) as file:
            yield file

    @contextmanager
    def ensure_open_bz2(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["rb"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[bz2.BZ2File, None, None]:
        """Ensure a BZ2-compressed file is downloaded and open a file inside it.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`bz2.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`bz2.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with bz2.open(path, **open_kwargs) as file:
            yield file

    def ensure_csv(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        read_csv_kwargs: Mapping[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Download a CSV and open as a dataframe with :mod:`pandas`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param read_csv_kwargs: Keyword arguments to pass through to
            :func:`pandas.read_csv`.

        :returns: A pandas DataFrame
        """
        import pandas as pd

        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return pd.read_csv(path, **_clean_csv_kwargs(read_csv_kwargs))

    def load_df(
        self,
        *subkeys: str,
        name: str,
        read_csv_kwargs: Mapping[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Open a pre-existing CSV as a dataframe with :mod:`pandas`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param read_csv_kwargs: Keyword arguments to pass through to
            :func:`pandas.read_csv`.

        :returns: A pandas DataFrame
        """
        import pandas as pd

        with self.open(*subkeys, name=name, mode="r", ensure_exists=False) as file:
            return pd.read_csv(file, **_clean_csv_kwargs(read_csv_kwargs))

    def dump_df(
        self,
        *subkeys: str,
        name: str,
        obj: pd.DataFrame,
        sep: str = "\t",
        index: bool = False,
        to_csv_kwargs: Mapping[str, Any] | None = None,
    ) -> None:
        """Dump a dataframe to a TSV file with :mod:`pandas`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param obj: The dataframe to dump
        :param sep: The separator to use, defaults to a tab
        :param index: Should the index be dumped? Defaults to false.
        :param to_csv_kwargs: Keyword arguments to pass through to
            :meth:`pandas.DataFrame.to_csv`.
        """
        to_csv_kwargs = {} if to_csv_kwargs is None else dict(to_csv_kwargs)
        to_csv_kwargs.setdefault("sep", sep)
        to_csv_kwargs.setdefault("index", index)
        # should this use unified opener instead? Pandas is pretty smart...
        path = self.join(*subkeys, name=name)
        obj.to_csv(path, **to_csv_kwargs)

    def ensure_json(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        open_kwargs: Mapping[str, Any] | None = None,
        json_load_kwargs: Mapping[str, Any] | None = None,
    ) -> JSON:
        """Download JSON and open with :mod:`json`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.

        :returns: A JSON object (list, dict, etc.)
        """
        with self.ensure_open(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            open_kwargs=open_kwargs,
        ) as file:
            return json.load(file, **(json_load_kwargs or {}))

    def ensure_json_bz2(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        open_kwargs: Mapping[str, Any] | None = None,
        json_load_kwargs: Mapping[str, Any] | None = None,
    ) -> JSON:
        """Download BZ2-compressed JSON and open with :mod:`json`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param open_kwargs: Additional keyword arguments passed to :func:`bz2.open`
        :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.

        :returns: A JSON object (list, dict, etc.)
        """
        with self.ensure_open_bz2(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            open_kwargs=open_kwargs,
        ) as file:
            return json.load(file, **(json_load_kwargs or {}))

    def ensure_yaml(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        open_kwargs: Mapping[str, Any] | None = None,
        yaml_load_kwargs: Mapping[str, Any] | None = None,
    ) -> JSON:
        """Download YAML and open with :mod:`yaml`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param yaml_load_kwargs: Keyword arguments to pass through to :func:`yaml.safe_load`.

        :returns: A JSON object (list, dict, etc.)
        """
        import yaml

        with self.ensure_open(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            open_kwargs=open_kwargs,
        ) as file:
            return yaml.safe_load(file, **(yaml_load_kwargs or {}))

    def load_yaml(
        self,
        *subkeys: str,
        name: str,
        open_kwargs: Mapping[str, Any] | None = None,
        yaml_load_kwargs: Mapping[str, Any] | None = None,
    ) -> JSON:
        """Open a JSON file :mod:`json`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param yaml_load_kwargs: Keyword arguments to pass through to :func:`yaml.safe_load`.

        :returns: A JSON object (list, dict, etc.)
        """
        import yaml

        with self.open(
            *subkeys, name=name, mode="r", open_kwargs=open_kwargs, ensure_exists=False
        ) as file:
            return yaml.safe_load(file, **(yaml_load_kwargs or {}))

    def load_json(
        self,
        *subkeys: str,
        name: str,
        open_kwargs: Mapping[str, Any] | None = None,
        json_load_kwargs: Mapping[str, Any] | None = None,
    ) -> JSON:
        """Open a JSON file :mod:`json`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.

        :returns: A JSON object (list, dict, etc.)
        """
        with self.open(
            *subkeys, name=name, mode="r", open_kwargs=open_kwargs, ensure_exists=False
        ) as file:
            return json.load(file, **(json_load_kwargs or {}))

    def dump_json(
        self,
        *subkeys: str,
        name: str,
        obj: JSON,
        open_kwargs: Mapping[str, Any] | None = None,
        json_dump_kwargs: Mapping[str, Any] | None = None,
    ) -> None:
        """Dump an object to a file with :mod:`json`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param json_dump_kwargs: Keyword arguments to pass through to :func:`json.dump`.
        """
        with self.open(
            *subkeys, name=name, mode="w", open_kwargs=open_kwargs, ensure_exists=True
        ) as file:
            json.dump(obj, file, **(json_dump_kwargs or {}))

    def ensure_pickle(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["rb"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
        pickle_load_kwargs: Mapping[str, Any] | None = None,
    ) -> Any:
        """Download a pickle file and open with :mod:`pickle`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param pickle_load_kwargs: Keyword arguments to pass through to
            :func:`pickle.load`.

        :returns: Any object
        """
        with self.ensure_open(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def load_pickle(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["rb"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
        pickle_load_kwargs: Mapping[str, Any] | None = None,
    ) -> Any:
        """Open a pickle file with :mod:`pickle`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param pickle_load_kwargs: Keyword arguments to pass through to
            :func:`pickle.load`.

        :returns: Any object
        """
        with self.open(
            *subkeys,
            name=name,
            mode=mode,
            open_kwargs=open_kwargs,
            ensure_exists=False,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def dump_pickle(
        self,
        *subkeys: str,
        name: str,
        obj: Any,
        mode: Literal["wb"] = "wb",
        open_kwargs: Mapping[str, Any] | None = None,
        pickle_dump_kwargs: Mapping[str, Any] | None = None,
    ) -> None:
        """Dump an object to a file with :mod:`pickle`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param pickle_dump_kwargs: Keyword arguments to pass through to
            :func:`pickle.dump`.
        """
        with self.open(
            *subkeys,
            name=name,
            mode=mode,
            open_kwargs=open_kwargs,
            ensure_exists=True,
        ) as file:
            pickle.dump(obj, file, **(pickle_dump_kwargs or {}))

    def ensure_pickle_gz(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        mode: Literal["rb"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
        pickle_load_kwargs: Mapping[str, Any] | None = None,
    ) -> Any:
        """Download a gzipped pickle file and open with :mod:`pickle`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`gzip.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
        :param pickle_load_kwargs: Keyword arguments to pass through to
            :func:`pickle.load`.

        :returns: Any object
        """
        with self.ensure_open_gz(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def load_pickle_gz(
        self,
        *subkeys: str,
        name: str,
        mode: Literal["rb"] = "rb",
        open_kwargs: Mapping[str, Any] | None = None,
        pickle_load_kwargs: Mapping[str, Any] | None = None,
    ) -> Any:
        """Open a gzipped pickle file with :mod:`pickle`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
        :param pickle_load_kwargs: Keyword arguments to pass through to
            :func:`pickle.load`.

        :returns: Any object
        """
        with self.open_gz(
            *subkeys,
            name=name,
            mode=mode,
            open_kwargs=open_kwargs,
            ensure_exists=False,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def ensure_excel(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        read_excel_kwargs: Mapping[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Download an excel file and open as a dataframe with :mod:`pandas`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param read_excel_kwargs: Keyword arguments to pass through to
            :func:`pandas.read_excel`.

        :returns: A pandas DataFrame
        """
        import pandas as pd

        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return pd.read_excel(path, **(read_excel_kwargs or {}))

    def ensure_tar_df(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        read_csv_kwargs: Mapping[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Download a tar file and open an inner file as a dataframe with :mod:`pandas`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param inner_path: The relative path to the file inside the archive
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param read_csv_kwargs: Keyword arguments to pass through to
            :func:`pandas.read_csv`.

        :returns: A dataframe

        .. warning::

            If you have lots of files to read in the same archive, it's better just to
            unzip first.
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_tarfile_csv(
            path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs)
        )

    def ensure_xml(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        parse_kwargs: Mapping[str, Any] | None = None,
    ) -> lxml.etree.ElementTree:
        """Download an XML file and open it with :mod:`lxml`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param parse_kwargs: Keyword arguments to pass through to
            :func:`lxml.etree.parse`.

        :returns: An ElementTree object

        .. warning::

            If you have lots of files to read in the same archive, it's better just to
            unzip first.
        """
        from lxml import etree

        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return etree.parse(path, **(parse_kwargs or {}))

    def load_xml(
        self,
        *subkeys: str,
        name: str,
        parse_kwargs: Mapping[str, Any] | None = None,
    ) -> lxml.etree.ElementTree:
        """Load an XML file with :mod:`lxml`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param parse_kwargs: Keyword arguments to pass through to
            :func:`lxml.etree.parse`.

        :returns: An ElementTree object

        .. warning::

            If you have lots of files to read in the same archive, it's better just to
            unzip first.
        """
        from lxml import etree

        with self.open(*subkeys, mode="r", name=name, ensure_exists=False) as file:
            return etree.parse(file, **(parse_kwargs or {}))

    def dump_xml(
        self,
        *subkeys: str,
        name: str,
        obj: lxml.etree.ElementTree,
        open_kwargs: Mapping[str, Any] | None = None,
        write_kwargs: Mapping[str, Any] | None = None,
    ) -> None:
        """Dump an XML element tree to a file with :mod:`lxml`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param write_kwargs: Keyword arguments to pass through to
            :func:`lxml.etree.ElementTree.write`.
        """
        with self.open(
            *subkeys, name=name, mode="wb", open_kwargs=open_kwargs, ensure_exists=True
        ) as file:
            obj.write(file, **(write_kwargs or {}))

    def ensure_tar_xml(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        parse_kwargs: Mapping[str, Any] | None = None,
    ) -> lxml.etree.ElementTree:
        """Download a tar file and open an inner file as an XML with :mod:`lxml`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param inner_path: The relative path to the file inside the archive
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param parse_kwargs: Keyword arguments to pass through to
            :func:`lxml.etree.parse`.

        :returns: An ElementTree object

        .. warning::

            If you have lots of files to read in the same archive, it's better just to
            unzip first.
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_tarfile_xml(path=path, inner_path=inner_path, **(parse_kwargs or {}))

    def ensure_zip_df(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        read_csv_kwargs: Mapping[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Download a zip file and open an inner file as a dataframe with :mod:`pandas`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param inner_path: The relative path to the file inside the archive
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param read_csv_kwargs: Keyword arguments to pass through to
            :func:`pandas.read_csv`.

        :returns: A pandas DataFrame
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_zipfile_csv(
            path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs)
        )

    def ensure_zip_np(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        load_kwargs: Mapping[str, Any] | None = None,
    ) -> numpy.typing.ArrayLike:
        """Download a zip file and open an inner file as an array-like with :mod:`numpy`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param inner_path: The relative path to the file inside the archive
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param load_kwargs: Additional keyword arguments that are passed through to
            :func:`read_zip_np` and transitively to :func:`numpy.load`.

        :returns: An array-like object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_zip_np(path=path, inner_path=inner_path, **(load_kwargs or {}))

    def ensure_rdf(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
        precache: bool = True,
        parse_kwargs: Mapping[str, Any] | None = None,
    ) -> rdflib.Graph:
        """Download a RDF file and open with :mod:`rdflib`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.
        :param precache: Should the parsed :class:`rdflib.Graph` be stored as a pickle
            for fast loading?
        :param parse_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.read_rdf` and transitively to
            :func:`rdflib.Graph.parse`.

        :returns: An RDF graph
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        if not precache:
            return read_rdf(path=path, **(parse_kwargs or {}))

        cache_path = path.with_suffix(path.suffix + ".pickle.gz")
        if cache_path.exists() and not force:
            import rdflib

            with gzip.open(cache_path, "rb") as file:
                return cast(rdflib.Graph, pickle.load(file))

        rv = read_rdf(path=path, **(parse_kwargs or {}))
        with gzip.open(cache_path, "wb") as file:
            pickle.dump(rv, file, protocol=pickle.HIGHEST_PROTOCOL)
        return rv

    def load_rdf(
        self,
        *subkeys: str,
        name: str | None = None,
        parse_kwargs: Mapping[str, Any] | None = None,
    ) -> rdflib.Graph:
        """Open an RDF file with :mod:`rdflib`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param parse_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.read_rdf` and transitively to
            :func:`rdflib.Graph.parse`.

        :returns: An RDF graph
        """
        path = self.join(*subkeys, name=name, ensure_exists=False)
        return read_rdf(path=path, **(parse_kwargs or {}))

    def dump_rdf(
        self,
        *subkeys: str,
        name: str,
        obj: rdflib.Graph,
        format: str = "turtle",
        serialize_kwargs: Mapping[str, Any] | None = None,
    ) -> None:
        """Dump an RDF graph to a file with :mod:`rdflib`.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param format: The format to dump in
        :param serialize_kwargs: Keyword arguments to through to
            :func:`rdflib.Graph.serialize`.
        """
        path = self.join(*subkeys, name=name, ensure_exists=False)
        serialize_kwargs = {} if serialize_kwargs is None else dict(serialize_kwargs)
        serialize_kwargs.setdefault("format", format)
        obj.serialize(path, **serialize_kwargs)

    def ensure_from_s3(
        self,
        *subkeys: str,
        s3_bucket: str,
        s3_key: str | Sequence[str],
        name: str | None = None,
        client: botocore.client.BaseClient | None = None,
        client_kwargs: Mapping[str, Any] | None = None,
        download_file_kwargs: Mapping[str, Any] | None = None,
        force: bool = False,
    ) -> Path:
        """Ensure a file is downloaded.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param s3_bucket: The S3 bucket name
        :param s3_key: The S3 key name
        :param name: Overrides the name of the file at the end of the S3 key, if given.
        :param client: A botocore client. If none given, one will be created
            automatically
        :param client_kwargs: Keyword arguments to be passed to the client on
            instantiation.
        :param download_file_kwargs: Keyword arguments to be passed to
            :func:`boto3.s3.transfer.S3Transfer.download_file`
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.

        :returns: The path of the file that has been downloaded (or already exists)
        """
        if not isinstance(s3_key, str):
            s3_key = "/".join(s3_key)  # join sequence
        if name is None:
            name = name_from_s3_key(s3_key)
        path = self.join(*subkeys, name=name, ensure_exists=True)
        download_from_s3(
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            path=path,
            client=client,
            client_kwargs=client_kwargs,
            force=force,
            download_file_kwargs=download_file_kwargs,
        )
        return path

    def ensure_from_google(
        self,
        *subkeys: str,
        name: str,
        file_id: str,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
    ) -> Path:
        """Ensure a file is downloaded from Google Drive.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file
        :param file_id: The file identifier of the Google file. If your share link is
            https://drive.google.com/file/d/1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z/view, then
            your file ID is ``1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z``.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download_from_google`.

        :returns: The path of the file that has been downloaded (or already exists)
        """
        path = self.join(*subkeys, name=name, ensure_exists=True)
        download_from_google(file_id, path, force=force, **(download_kwargs or {}))
        return path

    @contextmanager
    def ensure_open_sqlite(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[sqlite3.Connection, None, None]:
        """Ensure and connect to a SQLite database.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.

        :yields: An instance of :class:`sqlite3.Connection` from :func:`sqlite3.connect`

        Example usage: >>> import pystow >>> import pandas as pd >>> url =
        "https://s3.amazonaws.com/bbop-sqlite/obi.db" >>> sql = "SELECT * FROM
        entailed_edge LIMIT 10" >>> module = pystow.module("test") >>> with
        module.ensure_open_sqlite(url=url) as conn: >>> df = pd.read_sql(sql, conn)
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        with closing(sqlite3.connect(path.as_posix())) as conn:
            yield conn

    @contextmanager
    def ensure_open_sqlite_gz(
        self,
        *subkeys: str,
        url: str,
        name: str | None = None,
        force: bool = False,
        download_kwargs: Mapping[str, Any] | None = None,
    ) -> Generator[sqlite3.Connection, None, None]:
        """Ensure and connect to a SQLite database that's gzipped.

        Unfortunately, it's a paid feature to directly read gzipped sqlite files, so
        this automatically gunzips it first.

        :param subkeys: A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url: The URL to download.
        :param name: Overrides the name of the file at the end of the URL, if given.
            Also useful for URLs that don't have proper filenames with extensions.
        :param force: Should the download be done again, even if the path already
            exists? Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to
            :func:`pystow.utils.download`.

        :yields: An instance of :class:`sqlite3.Connection` from :func:`sqlite3.connect`

        Example usage: >>> import pystow >>> import pandas as pd >>> url =
        "https://s3.amazonaws.com/bbop-sqlite/hp.db.gz" >>> module =
        pystow.module("test") >>> sql = "SELECT * FROM entailed_edge LIMIT 10" >>> with
        module.ensure_open_sqlite_gz(url=url) as conn: >>> df = pd.read_sql(sql, conn)
        """
        path = self.ensure_gunzip(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        with closing(sqlite3.connect(path.as_posix())) as conn:
            yield conn


def _clean_csv_kwargs(read_csv_kwargs: None | Mapping[str, Any]) -> dict[str, Any]:
    read_csv_kwargs = {} if read_csv_kwargs is None else dict(read_csv_kwargs)
    read_csv_kwargs.setdefault("sep", "\t")
    return read_csv_kwargs
